class_name State
extends Node
## A data-only state for Tic-Tac-Toe.
##
## Illustrate the separation of the game logic from the user interface.
##

## The size of the game grid.
@export var size: int = 3 # visible in Godot's Inspector

## Player/board piece marks.
enum Player {
	Cross = 1, 		# Player Cross
	Circle = -1, 	# Player Circle
	Empty = 0	# Empty position
}

## The active player.
var player: Player = Player.Empty

## The board.
## An array size x size
var board: Array = []
	
## The next player of p
static func next(p:Player) -> Player:
	match p:
		Player.Cross:
			return Player.Circle
		Player.Circle:
			return Player.Cross
		_:
			return Player.Empty

## Player string format
static func player_to_string(p:Player) -> String:
	var s:String = " " # Default: Empty
	match p:
		Player.Cross:
			s = "X"
		Player.Circle:
			s = "O"
		_:
			s = " "
	return s

## Board initialization: empty n x n grid
static func new_board(n:int) -> Array:
	var b = [];
	for i in range(n): # i - rows
		var r: Array[Player] = [];
		r.resize(n)
		r.fill(Player.Empty)	# Using Array.fill
		b.append(r)			# Array.fill doesn't apply to Variants
	return b
	
## State construction from player and board.
## The deep parameter controls the copy mode (deep/shallow) of the board.
static func new_state(p:Player, b:Array, deep:bool = false) -> State:
	var s = State.new()
	s.size = b.size()
	s.board = b.duplicate(deep)
	s.player = p	
	return s

## Print (text mode)
static func state_to_string(s:State) -> String:
	var hline = "---".join(range(s.board.size()).map(func(_x): return "-"))
	var strs: Array[String] = ["Player: %s" % player_to_string(s.player), hline]
	for r in range(s.size):
		var row_str: Array = s.board[r].map(player_to_string)
		strs.append(" | ".join(row_str))
		strs.append(hline)
	strs.append(hline)
	return "\n".join(strs)
			
## Board column j
## j is checked
func col(j:int) -> Array[Player]:
	var index_ok = 0 <= j and j < size
	if !index_ok:
		return []
	var c: Array[Player] = []
	for i in range(size):
		c.append(board[i][j])
	return c
			
## Board row i
## i is checked
func row(i:int) -> Array[Player]:
	var index_ok = 0 <= i and i < size
	if !index_ok:
		return []
	var r: Array[Player] = []
	for j in range(size):
		r.append(board[i][j])
	return r
	
## Board descending diagonal
func dd()  -> Array:
	return range(size).map(func (i): return board[i][i])
	
## Board ascending diagonal
func ad()  -> Array:
	return range(size).map(func (i): return board[i][size - 1 - i])
	
# State view for player p
func view(_p:Player) -> State:
	return new_state(player, board, true) # Deep copy
	
## State constructor
func _init(_size:int = 3):
	self.size = _size
	setup()

## Initial state
func setup():
	board = new_board(size)
	player = Player.Cross

## Action: player p places a piece in ij
func action(p:Player, i:int, j:int) -> State:
	var index_ok:bool = 0 <= i and i < size and \
		0 <= j and j < size
	var action_ok:bool = p == player and \
		index_ok and \
		board[i][j] == Player.Empty
	if !action_ok:
		return self
	# We're good to continue
	var next_p = next(player)
	var next_b = board.duplicate(true)
	next_b[i][j] = player
	return new_state(next_p, next_b)
	
## Victory condition for player p
func victory(p:Player) -> bool:
	var is_p = func(x): return x == p
	var condition:bool = p != Player.Empty and (
		dd().all(is_p) or ad().all(is_p) or
		range(size).any( func (i): return \
			col(i).all(is_p) or row(i).all(is_p)
		))
	return condition

## Draw state
func draw() -> bool:
	return !victory(Player.Cross) and !victory(Player.Circle) and range(size).all( \
		func (r): return !board[r].has(Player.Empty))
		
func game_end() -> bool:
	return victory(Player.Cross) or victory(Player.Circle) or draw()
		
## Progress of play
func progress(p:Player, i:int, j:int) -> State:
	if !game_end():
		return action(p, i, j)
	else:
		player = Player.Empty
		return self 
		
func step(i, j) -> void:
	var next_state:State = progress(player, i, j)
	player = next_state.player
	board = next_state.board.duplicate(true)
		
## Called when the node enters the scene tree for the first time.
#func _ready() -> void:
	#setup()
#
## Called every frame. 'delta' is the elapsed time since the previous frame.
#func _process(_delta: float) -> void:
	#var i:int = randi() % size
	#var j:int = randi() % size
	#step(i, j)
	#var v  = view(player)
	#print(state_to_string(v))
